import matplotlib

matplotlib.use("agg")
from collections import OrderedDict
import contextlib
import io
from matplotlib import pyplot as plt
import numpy as np
import os
import pyperf
import subprocess
import sys


@contextlib.contextmanager
def stdoutIO(stdout=None):
    old = sys.stdout
    if stdout is None:
        stdout = io.StringIO()
    sys.stdout = stdout
    yield stdout
    sys.stdout = old


def run_perf(args, json_name):
    if os.path.exists(json_name):
        return
    args = args + ["-o", json_name]
    print(args)
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    print(out.decode())
    print(err.decode())


def make_plot(extension):
    ratios = OrderedDict()
    stddevs = OrderedDict()
    benchmarks = OrderedDict()
    np_bench = pyperf.Benchmark.load(open("{}_{}".format("numpy", extension), "r"))
    np_mean = np_bench.mean()
    np_stddev = np_bench.stdev()
    for package in setup:
        if package == "numpy":
            continue
        benchmarks[package] = pyperf.Benchmark.load(
            open("{}_{}".format(package, extension), "r")
        )
        mean = benchmarks[package].mean()
        stddev = benchmarks[package].stdev()
        ratios[package] = mean / np_mean
        stddevs[package] = ratios[package] * np.sqrt(
            (np_stddev / np_mean) ** 2 + (stddev / mean) ** 2
        )
    fig, ax = plt.subplots()
    packages = list(ratios.keys())
    ax.bar(packages, ratios.values(), yerr=stddevs.values())
    fig.suptitle(extension.replace(".json", "").replace("_", " ").title())
    ax.set_ylabel("numpy overhead (x time for numpy); lower is better")
    plt.savefig(extension.replace(".json", ".png"))
    plt.close(fig)

    if ratios["unyt"] != min(ratios.values()):
        rvalues = list(ratios.values())
        svalues = list(stddevs.values())
        unyt_index = packages.index("unyt")
        min_index = rvalues.index(min(rvalues))
        if ratios["unyt"] > 3 * svalues[min_index] + rvalues[min_index]:
            for package in ratios:
                script = get_script(benchmarks, package)
                with stdoutIO() as s:
                    exec(script)
                res = s.getvalue().replace("\n", "")
                print(
                    "{}: {} +- {} ({})".format(
                        package, ratios[package], stddevs[package], res
                    )
                )
            print(get_script(benchmarks, "unyt"))


def get_script(benchmarks, package):
    meta = benchmarks[package].get_metadata()
    setup_s = meta["timeit_setup"][1:-1]
    bench_s = "print(" + meta["timeit_stmt"][1:-1] + ")"
    script = setup_s + "; " + bench_s
    script = script.replace("; ", "\n")
    return script


setup = OrderedDict(
    [
        ("numpy", "import numpy as np"),
        ("pint", "from pint import UnitRegistry; u = UnitRegistry()"),
        ("astropy", "import astropy.units as u"),
        ("unyt", "import unyt as u"),
        ("quantities", "import quantities as u"),
    ]
)

base_args = ["python", "-m", "pyperf", "timeit"]

shared_setup = "import numpy as np; import operator"

base_setups = OrderedDict(
    [
        ("small_list", "data = [1., 2., 3.]"),
        ("small_tuple", "data = (1., 2., 3.)"),
        ("small_array", "data = np.array([1., 2., 3.])"),
        ("big_list", "data = (np.arange(1e6)+1).tolist()"),
        ("big_array", "data = (np.arange(1e6)+1)"),
    ]
)

op_ufuncs = OrderedDict(
    [
        ("operator.add", "np.add"),
        ("operator.sub", "np.subtract"),
        ("operator.mul", "np.multiply"),
        ("operator.truediv", "np.true_divide"),
        ("operator.eq", "np.equal"),
    ]
)

for bs in base_setups:
    for package in sorted(setup):
        print(package)
        setup_s = "; ".join([shared_setup, setup[package], base_setups[bs]])
        args = base_args + ["-s", setup_s + " "]
        if package == "numpy":
            args.append("np.array(data)")
        else:
            args.append("data*u.g")
        json_name = "{}_{}_create.json".format(package, bs)
        run_perf(args, json_name)

        if "list" in bs or "tuple" in bs:
            continue

        args = base_args + ["-s", setup_s + "; data=np.asarray(data); out=data.copy()"]
        if package == "numpy":
            args[-1] += "; "
        else:
            if package != "pint":
                args[-1] += "*u.g"
            args[-1] += "; data = data*u.g "

        args.append("data**2")
        json_name = "{}_{}_square.json".format(package, bs)
        run_perf(args, json_name)

        args[-1] = "np.power(data, 2)"
        json_name = "{}_{}_npsquare.json".format(package, bs)
        run_perf(args, json_name)

        args[-1] = "np.power(data, 2, out=out)"
        json_name = "{}_{}_npsquareout.json".format(package, bs)
        run_perf(args, json_name)

        args[-1] = "data**0.5"
        json_name = "{}_{}_sqrt.json".format(package, bs)
        run_perf(args, json_name)

        args[-1] = "np.sqrt(data)"
        json_name = "{}_{}_npsqrt.json".format(package, bs)
        run_perf(args, json_name)

        args[-1] = "np.sqrt(data, out=out)"
        json_name = "{}_{}_npsqrtout.json".format(package, bs)
        run_perf(args, json_name)

    make_plot("{}_create.json".format(bs))
    if "list" not in bs and "tuple" not in bs:
        make_plot("{}_square.json".format(bs))
        make_plot("{}_npsquare.json".format(bs))
        make_plot("{}_npsquareout.json".format(bs))
        make_plot("{}_sqrt.json".format(bs))
        make_plot("{}_npsqrt.json".format(bs))
        make_plot("{}_npsqrtout.json".format(bs))


for bs in base_setups:
    if "list" in bs or "tuple" in bs:
        continue
    for op, ufunc in op_ufuncs.items():
        for bench, bench_name in [
            (op + r"(data1, data2)", op + "12.json"),
            (op + r"(data2, data1)", op + "21.json"),
            (ufunc + r"(data1, data2)", ufunc + "12.json"),
            (ufunc + r"(data2, data1)", ufunc + "21.json"),
            (ufunc + r"(data1, data2, out=out)", ufunc + "12out.json"),
            (ufunc + r"(data2, data1, out=out)", ufunc + "21out.json"),
        ]:
            for unit_choice in [("g", "g"), ("kg", "g")]:
                for package in sorted(setup):
                    print(package)
                    setup_s = (
                        "; ".join([shared_setup, setup[package], base_setups[bs]])
                        + "; "
                    )
                    if "out" in bench:
                        if package not in ("pint", "numpy") and "equal" not in bench:
                            setup_s += "out=data*u.{}; ".format(unit_choice[0])
                        else:
                            setup_s += "out=np.array(data); "
                    if package == "numpy":
                        setup_s += "; ".join(
                            [r"data1 = np.array(data)", r"data2 = np.array(data)"]
                        )
                        if unit_choice[0] != unit_choice[1]:
                            _bench = bench.replace("data1", ".001*data1")
                    else:
                        setup_s += "; ".join(
                            [
                                "data1 = data*u.{}".format(unit_choice[0]),
                                "data2 = data*u.{}".format(unit_choice[1]),
                            ]
                        )
                        _bench = bench
                    args = base_args + ["-s", setup_s + " "]
                    json_name = "{}_{}_{}{}".format(
                        package, bs, unit_choice[0], unit_choice[1]
                    )
                    run_perf(args + [_bench], json_name + "_" + bench_name)
                make_plot(
                    "{}_{}{}_{}".format(bs, unit_choice[0], unit_choice[1], bench_name)
                )



"""The setup script."""

import os
import sys

from setuptools import find_packages, setup

# ensure the current directory is on sys.path
# so versioneer can be imported when pip uses
# PEP 517/518 build rules
sys.path.append(os.path.dirname(__file__))

import versioneer  # NOQA

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["numpy>=1.17.5", "sympy>=1.5"]

test_requirements = ["pytest"]

setup(
    author="The yt project",
    author_email="yt-dev@python.org",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="A package for handling numpy arrays with units",
    install_requires=requirements,
    license="BSD license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    package_data={"unyt": ["tests/data/old_json_registry.txt"]},
    keywords="unyt",
    name="unyt",
    packages=find_packages(include=["unyt", "unyt.tests", "unyt.tests.data"]),
    test_suite="tests",
    tests_require=test_requirements,
    python_requires=">=3.8",
    url="https://github.com/yt-project/unyt",
    version=versioneer.get_version(),
    zip_safe=False,
    cmdclass=versioneer.get_cmdclass(),
)



from docutils.parsers.rst import Directive

try:
    maketrans = "".maketrans
except AttributeError:
    # python2 fallback
    from string import maketrans

import numpy as np
from sympy.core.mul import Mul
from sympy.core.symbol import Symbol

import unyt
import unyt.dimensions as dims
from unyt import Unit
from unyt._unit_lookup_table import name_alternatives, physical_constants
from unyt.exceptions import UnitsNotReducible
from unyt.unit_registry import default_unit_registry
from unyt.unit_systems import _split_prefix

all_dims = {}
for d in dims.__dict__.keys():
    if isinstance(getattr(dims, d), (Symbol, Mul)):
        if d == "dimensionless":
            continue
        all_dims[getattr(dims, d)] = d


def setup(app):
    app.add_directive("show_all_units", ShowAllUnits)
    app.add_directive("show_all_constants", ShowAllConstants)
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    retdict = dict(version="0.1")

    return retdict


class ShowAllUnits(Directive):
    required_arguments = 0
    optional_arguments = 0

    def run(self):
        lines = []

        for name, alt_names in name_alternatives.items():
            prefix, base = _split_prefix(name, default_unit_registry.lut)
            if prefix != "":
                continue
            lut_entry = default_unit_registry[name]
            u = Unit(name)
            try:
                dimensions = all_dims[u.dimensions]
            except KeyError:
                if u.is_dimensionless:
                    dimensions = "dimensionless"
                else:
                    dimensions = u.dimensions
            try:
                mks_value = (1 * u).in_mks()
            except UnitsNotReducible:
                mks_value = "N/A"
            try:
                cgs_value = (1 * u).in_cgs()
            except UnitsNotReducible:
                cgs_value = "N/A"

            def generate_table_value(value):
                if value == "N/A":
                    return value
                approx_string = "{:.4e}".format(value)
                real_string = str(value)
                fv = value.value
                close_value = float("{:.4e}".format(fv))

                if (close_value - fv) / fv < 1e-6 and len(str(fv)) > 8:
                    return approx_string

                if fv < 1e-4 or fv > 1e4 or len(str(fv)) > 8:
                    return approx_string

                return real_string

            latex_repr = "``" + u.latex_repr + "``"
            if latex_repr == "````":
                latex_repr = ""
            with np.printoptions(precision=4, suppress=False, floatmode="maxprec"):
                lines.append(
                    (
                        name,
                        str(dimensions),
                        generate_table_value(mks_value),
                        generate_table_value(cgs_value),
                        latex_repr,
                        str(lut_entry[4]),
                        ", ".join([a for a in alt_names if a != name]),
                    )
                )
        lines.insert(
            0,
            [
                "Unit Name",
                "Dimensions",
                "MKS value",
                "CGS Value",
                "LaTeX Representation",
                "SI Prefixable?",
                "Alternate Names",
            ],
        )
        lines = as_rest_table(lines, full=False).split("\n")
        rst_file = self.state_machine.document.attributes["source"]
        self.state_machine.insert_input(lines, rst_file)
        return []


class ShowAllConstants(Directive):
    required_arguments = 0
    optional_arguments = 0

    def run(self):
        lines = []

        for name, (_value, _unit, alternate_names) in physical_constants.items():
            val = getattr(unyt.physical_constants, name)
            if val > 1e4 or val < 1e-4:
                default_value = "{:.4e}".format(val)
            else:
                default_value = str(val)
            lines.append((name, default_value, ", ".join(alternate_names)))

        lines.insert(0, ["Constant Name", "Value", "Alternate Names"])
        lines = as_rest_table(lines, full=False).split("\n")
        rst_file = self.state_machine.document.attributes["source"]
        self.state_machine.insert_input(lines, rst_file)
        return []


def as_rest_table(data, full=False):
    """
    Originally from ActiveState recipes, copy/pasted from GitHub
    where it is listed with an MIT license.

    https://github.com/ActiveState/code/tree/master/recipes/Python/579054_Generate_Sphinx_table

    """
    data = data if data else [["No Data"]]
    table = []
    # max size of each column
    sizes = list(map(max, zip(*[[len(str(elt)) for elt in member] for member in data])))
    num_elts = len(sizes)

    if full:
        start_of_line = "| "
        vertical_separator = " | "
        end_of_line = " |"
        line_marker = "-"
    else:
        start_of_line = ""
        vertical_separator = "  "
        end_of_line = ""
        line_marker = "="

    meta_template = vertical_separator.join(
        ["{{{{{0}:{{{0}}}}}}}".format(i) for i in range(num_elts)]
    )
    template = "{0}{1}{2}".format(
        start_of_line, meta_template.format(*sizes), end_of_line
    )
    # determine top/bottom borders
    if full:
        to_separator = maketrans("| ", "+-")
    else:
        to_separator = maketrans("|", "+")
    start_of_line = start_of_line.translate(to_separator)
    vertical_separator = vertical_separator.translate(to_separator)
    end_of_line = end_of_line.translate(to_separator)
    separator = "{0}{1}{2}".format(
        start_of_line,
        vertical_separator.join([x * line_marker for x in sizes]),
        end_of_line,
    )
    # determine header separator
    th_separator_tr = maketrans("-", "=")
    start_of_line = start_of_line.translate(th_separator_tr)
    line_marker = line_marker.translate(th_separator_tr)
    vertical_separator = vertical_separator.translate(th_separator_tr)
    end_of_line = end_of_line.translate(th_separator_tr)
    th_separator = "{0}{1}{2}".format(
        start_of_line,
        vertical_separator.join([x * line_marker for x in sizes]),
        end_of_line,
    )
    # prepare result
    table.append(separator)
    # set table header
    titles = data[0]
    table.append(template.format(*titles))
    table.append(th_separator)

    for d in data[1:-1]:
        table.append(template.format(*d))
        if full:
            table.append(separator)
    table.append(template.format(*data[-1]))
    table.append(separator)

    return "\n".join(table)



from collections import OrderedDict

import numpy as np
import perf
from matplotlib import pyplot as plt
from matplotlib.patches import Patch

ALPHA_MAP = {"small": 0.3333, "medium": 0.666, "big": 1.0}
COLOR_MAP = {"unyt": "C0", "astropy": "C1", "pint": "C2"}
SIZE_LABELS = {"small": "3", "medium": "$10^3$", "big": "$10^6$"}


def make_plot(benchmark_name, benchmarks, fig_filename):
    plt.style.use("tableau-colorblind10")
    plt.rc("font", family="stixgeneral")
    plt.rc("mathtext", fontset="cm")
    fig, ax = plt.subplots()

    ratios = OrderedDict()
    stddevs = OrderedDict()

    width = 0.1
    packages = ["pint", "astropy", "unyt"]
    package_offsets = [-width, 0, width]
    size_offsets = [-width * 3, 0, width * 3]
    static_offset = 0
    sizes = ["big", "medium", "small"]
    yticks = []
    all_yticks = []

    for ind, benchmark in enumerate(benchmarks):
        if "create" not in benchmark:
            benchmark = "array_" + benchmark
        for size, size_offset in zip(sizes, size_offsets):
            np_fname = "../benchmarks/numpy_{}_{}.json".format(size, benchmark)
            with open(np_fname, "r") as f:
                np_bench = perf.Benchmark.load(f)
            np_mean = np_bench.mean()
            np_stddev = np_bench.stdev()
            for package, package_offset in zip(packages, package_offsets):
                fname = "../benchmarks/{}_{}_{}.json".format(package, size, benchmark)
                with open(fname, "r") as f:
                    pbench = perf.Benchmark.load(f)
                mean = pbench.mean()
                stddev = pbench.stdev()
                ratios[package] = mean / np_mean
                stddevs[package] = ratios[package] * np.sqrt(
                    (np_stddev / np_mean) ** 2 + (stddev / mean) ** 2
                )
                ytick = ind + 1 + size_offset + package_offset + static_offset
                ax.barh(
                    ytick,
                    ratios[package],
                    width,
                    xerr=stddevs[package],
                    color=COLOR_MAP[package],
                    label=" ".join([package, SIZE_LABELS[size]]),
                    alpha=ALPHA_MAP[size],
                )
                all_yticks.append(ytick)
                if package == "astropy" and size == "medium":
                    yticks.append(ytick)
            static_offset += 0.01

    legend_patches = [Patch(color=c) for c in COLOR_MAP.values()]
    leg = ax.legend(legend_patches, COLOR_MAP.keys(), loc=1)
    size_patches = [Patch(color="k", alpha=a) for a in ALPHA_MAP.values()]
    size_labels = [SIZE_LABELS[l] for l in ALPHA_MAP.keys()]
    ax.legend(size_patches, size_labels, loc=4, title="Number of elements")
    ax.add_artist(leg)
    ax.set_xlabel(r"$T_{\rm package} / T_{\rm numpy}$")
    ax.set_xscale("symlog", linthresh=1)
    ax.set_xlim(0, 1000)
    ax.set_yticks(yticks)
    ax.set_xticks(
        [
            0.1,
            0.2,
            0.3,
            0.4,
            0.5,
            0.6,
            0.7,
            0.8,
            0.9,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            200,
            300,
            400,
            500,
            600,
            700,
            800,
            900,
        ],
        minor=True,
    )
    ax.set_xticks([0, 1, 10, 100, 1000])
    ax.set_xticklabels(["0", "1", "10", "100", "1000"])
    ax.set_yticklabels(benchmarks.values())
    ax.plot([1, 1], [0, len(benchmarks) + 1], "--", color="k", lw=0.75, alpha=0.5)
    spacing = len(benchmarks) // 2 + 1
    ax.set_ylim(all_yticks[0] - 0.05 * spacing, all_yticks[-1] + 0.05 * spacing)
    fig.suptitle(benchmark_name)
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.savefig(fig_filename, dpi=300)
    plt.close(fig)


make_plot(
    benchmark_name="Apply units to data",
    benchmarks={"list_create": "List", "array_create": "Array"},
    fig_filename="apply.png",
)

make_plot(
    benchmark_name="Unary operations",
    benchmarks={"sqrt": r"$\mathtt{data**0.5}$", "square": r"$\mathtt{data**2}$"},
    fig_filename="unary.png",
)

make_plot(
    benchmark_name="Binary operations, different units",
    benchmarks={
        "kgg_operator.add12": r"$\mathtt{a + b}$",
        "kgg_operator.sub12": r"$\mathtt{a - b}$",
        "kgg_operator.mul12": r"$\mathtt{a * b}$",
        "kgg_operator.truediv12": r"$\mathtt{a / b}$",
        "kgg_operator.eq12": r"$\mathtt{a == b}$",
    },
    fig_filename="binary_different_units.png",
)

make_plot(
    benchmark_name="Binary operations, same units",
    benchmarks={
        "gg_operator.add12": r"$\mathtt{a + b}$",
        "gg_operator.sub12": r"$\mathtt{a - b}$",
        "gg_operator.mul12": r"$\mathtt{a * b}$",
        "gg_operator.truediv12": r"$\mathtt{a / b}$",
        "gg_operator.eq12": r"$\mathtt{a == b}$",
    },
    fig_filename="binary_same_units.png",
)

make_plot(
    benchmark_name="NumPy ufunc",
    benchmarks={
        "kgg_np.add12": r"$\mathtt{np.add(a, b)}$",
        "kgg_np.subtract12": r"$\mathtt{np.subtract(a, b)}$",
        "kgg_np.multiply12": r"$\mathtt{np.multiply(a, b)}$",
        "kgg_np.true_divide12": r"$\mathtt{np.divide(a, b)}$",
        "kgg_np.equal12": r"$\mathtt{np.equal(a, b)}$",
        "npsqrt": r"$\mathtt{np.sqrt(data)}$",
        "npsquare": r"$\mathtt{np.power(data, 2)}$",
    },
    fig_filename="ufunc.png",
)

make_plot(
    benchmark_name="In-place ufunc",
    benchmarks={
        "kgg_np.add12out": r"$\mathtt{np.add(a, b, out=out)}$",
        "kgg_np.subtract12out": r"$\mathtt{np.subtract(a, b, out=out)}$",
        "kgg_np.multiply12out": r"$\mathtt{np.multiply(a, b, out=out)}$",
        "kgg_np.true_divide12out": r"$\mathtt{np.divide(a, b, out=out)}$",
        "kgg_np.equal12out": r"$\mathtt{np.equal(a, b, out=out)}$",
        "npsqrtout": r"$\mathtt{np.sqrt(data, out=out)}$",
        "npsquareout": r"$\mathtt{np.power(data, 2, out=out)}$",
    },
    fig_filename="ufuncout.png",
)
